#include "AudioChainElement.h"
#include <iostream>

AudioChainElement::AudioChainElement(void)
{
  DEBUG_PRINTF("AudioChainElement constructor\n");
  // Initialize the input and output to zero. 
  // Child classes can override this behavior in their constructors.
  m_nextTickInput = m_currentOutput = 0.0;
  m_visited = false;
  m_tickDone = false;
  m_gain = 1.0;
}

AudioChainElement::~AudioChainElement(void)
{
  // The output list will recursively delete itself. 
}

SAMPLE AudioChainElement::getOutput(void)
{
  MyFloat64 returnOutput;
  // make sure the output state is defined - i.e. make sure we have 
  // already executed this tick. (or should we just sustain the old output?)
  assert(m_visited);

  returnOutput =  m_gain*m_currentOutput;

  if((returnOutput > 1.0) || (returnOutput < -1.0)) {
    printf("Output out of range %f\n",returnOutput);
  }

  
  return returnOutput;
}

void AudioChainElement::tick(SAMPLE input)
{
  SAMPLE output = 0;
  SAMPLE currentInput;  
  ListItem* pItem;
  AudioChainElement* pAudioElement;

  // tick() has been already been called for the current sample
  // this is the recursion base case
  if (m_visited) {
    // accumulate all inputs
    m_nextTickInput += input;
    return;
  }
  else {
    m_visited = true;
    m_tickDone = false;
    // Consider the input to be the sum of the input sample given as an agrument
    // and all inputs from the other input sources (accumulated in m_nextTickInput).
    currentInput = input + m_nextTickInput;
    // Reset input for next sample
    m_nextTickInput = 0;
    
    // This is where the sound actually gets generated
    // Virtual function call - descendent classes define
    // their own algorithms for generating output samples.
    m_currentOutput = output = generateOutput(currentInput);
    if((output > 1.0) || (output < -1.0)) {
      printf("Output out of range %f\n",output);
    }
    
    // Output this element's sound to all the audio elements it is 
    // connected to.
    pItem = m_outputList.getHead();
    while (pItem != NULL) {
      pAudioElement = (AudioChainElement *) pItem->getData();
      assert(pAudioElement != NULL);
      
      // Recursive step
      pAudioElement->tick(m_gain*output);
      
      pItem = pItem->getNext();
    }
  }
}

// Notify the element and all those linked to it that the processing of the current 
// sample is done.
void AudioChainElement::endOfTick(void)
{
  ListItem* pItem = NULL;
  // recursion base case - this element already knows the tick is done
  if (m_tickDone) {
    return;
  }

  m_visited = false;
  m_tickDone = true;

  // Recursively notify other elements of the end of tick.
  pItem = m_outputList.getHead();
  while (pItem != NULL) {
    AudioChainElement * pAudioElement = (AudioChainElement *) pItem->getData();
    assert(pAudioElement != NULL);

    // Recursive step
    pAudioElement->endOfTick();

    pItem = pItem->getNext();
  }
}



void AudioChainElement::connectTo(AudioChainElement* element)
{
  // Add the element to our output list
  m_outputList.insert(element);  
}
