#include "CircularBuffer.h"
#include "assert.h"
#include "stdio.h"
#include "PolyCommon.h"


template <class Element>
CircularBuffer<Element>::CircularBuffer(int maxCapacity, int delay, bool initBuffer, Element initValue)
{
  m_Buffer = new Element[maxCapacity];
  // Sanity checks
  assert((void*)m_Buffer != NULL);
  assert(delay < maxCapacity);
  assert(maxCapacity > 0);
  assert(delay >=0);

  if (initBuffer) {
    // Initialize buffer
    for (int i=0; i<maxCapacity; i++) {
      m_Buffer[i] = initValue;
    }
  }

  m_count = 0;
  m_maxCapacity = maxCapacity;
  m_delay = delay;

  // Set up write index
  m_writeIndex = 0;

  // Set up read index to point 'delay' slots behind the write index
  m_readIndex = m_writeIndex - delay;
  if (m_readIndex < 0 ) {
    m_readIndex += maxCapacity;
  }
  DEBUG_PRINTF("CircularBuffer constructor: writeIndex = %d, m_readIndex = %d\n",m_writeIndex, m_readIndex);
}

template <class Element>
CircularBuffer<Element>::~CircularBuffer(void)
{
  delete[] m_Buffer;
}

template <class Element>
Element CircularBuffer<Element>::pushPop(Element pushedElement)
{
  // Note: must do read before the write as both r/w pointers may
  // be pointing to the same slot.

  /************** Read **********************/
  Element poppedElement = m_Buffer[m_readIndex];
  // Increment read index
  m_readIndex++;
  // Wrap the index at the end of the array
  if (m_readIndex >= m_maxCapacity) {
    m_readIndex = 0;
  }

  /************** Write **********************/

  m_Buffer[m_writeIndex] = pushedElement;
 
  // Increment write index
  m_writeIndex++;
  // Wrap the index at the end of the array
  if (m_writeIndex >= m_maxCapacity) {
    m_writeIndex = 0;
  }

  return poppedElement;
}

template <class Element>
Element CircularBuffer<Element>::peekBack(int numSlots)
{
  int index;
  assert(numSlots < maxCapacity);
  
  index = m_writeIndex-1-numSlots;
  if (index < 0) {
    index+= m_maxCapacity;
  }
  return m_Buffer[index];
}

template <class Element>
bool CircularBuffer<Element>::setDelay(int delay)
{
  if(delay >= maxCapacity) {
    return false;
  }

  // Set up read index to point 'delay' slots behind the write index
  m_readIndex = m_writeIndex - delay;
  if(readIndex < 0 ) {
    m_readIndex += maxCapacity;
  }

  m_Delay = delay;

  DEBUG_PRINTF("CircularBuffer setDelay: new delay = %d"
               "m_writeIndex = %d, m_readIndex = %d\n", m_Delay, m_writeIndex, m_readIndex);
}


template <class Element>
void CircularBuffer<Element>::dump(void)
{
  printf("Dumping content of buffer:\n");
  if(m_Buffer) {
    for(int i=0; i<m_maxCapacity; i++) {
      printf("Buffer[%d] = %f\n",i,(double)m_Buffer[i]);
    }
  }

  printf("m_dealy = %d\n",m_delay);
  printf("m_maxCapacity = %d\n",m_maxCapacity);
  //printf("m_count = %d\n",m_count);
  printf("m_writeIndex = %d\n",m_writeIndex);
  printf("m_readIndex = %d\n",m_readIndex);
}



//#define BUFFER_UNIT_TEST
#ifdef BUFFER_UNIT_TEST

int main(int argc, char** argv)
{
  CircularBuffer<double> buffer(20,5,0);
  double input[8]={1,2,3,4,5,6,7,8};
  double output[8];
  //write
  buffer.write(input,8);
  printf("State of buffer after write:\n");
  printf("-----------------------------\n");
  buffer.dump();
  
  //read
  buffer.read(output,8);
  printf("State of buffer after read:\n");
  printf("-----------------------------\n");
  buffer.dump();

  //examine output
  printf("\n\n\nDumping output:\n");
  for(int i=0; i<8; i++) {
    printf("output[%d] = %f\n",i,output[i]);
  }
  return 0;

}

#endif //BUFFER_UNIT_TEST







/*
template <class Element>
int CircularBuffer<Element>::write(const Element* inputBuffer, int numElements)
{
  int i=0;
  for (i=0; i<numElements; i++) {
    if (m_count < m_maxCapacity) {
      m_Buffer[m_writeIndex] = inputBuffer[i];    
      m_count++;

      // Increment write index
      m_writeIndex++;
      // Wrap the index at the end of the array
      if (m_writeIndex >= m_maxCapacity) {
        m_writeIndex = 0;
      }
    }
    // Overflow!!
    else {
      break;
    }
  }
  // Return number of elements successfully written
  return i;
}

template <class Element>
int CircularBuffer<Element>::read(Element* outputBuffer, int numElements)
{
  int i=0;
  for (i=0; i<numElements; i++) {
    // Make sure buffer is not empty
    if (m_count > 0) {
      outputBuffer[i] = m_Buffer[m_readIndex];
      m_count--;
      
      // Increment read index
      m_readIndex++;
      // Wrap the index at the end of the array
      if (m_readIndex >= m_maxCapacity) {
        m_readIndex = 0;
      }
    }
    // Underflow!!
    else {
      break;
    }
  }
  // Return number of elements successfully read
  return i;
}
*/
