#include "PluckedString.h"
#include <math.h>

#define DECAY_FEEDBACK_GAIN    (0.995)
#define RELEASE_FEEDBACK_GAIN  (0.85)
#define FINISH_GAIN_THRESH     (0.0001)
#define ROUND__(_x) floor(_x + 0.5)



const MyFloat64 PluckedString::baseNumReleaseSamples = log(FINISH_GAIN_THRESH) / log (RELEASE_FEEDBACK_GAIN);

PluckedString::PluckedString(MyFloat64 freq, unsigned int samplingRate, 
                             ReleaseCallback callback, unsigned int callbackUserData)
{
  int delayLength, maxDelayLength;
  assert(freq > MIN_STRING_FREQ);

  delayLength = frequencyToDelayLength(freq, samplingRate);
  maxDelayLength = frequencyToDelayLength(MIN_STRING_FREQ, samplingRate);

  m_pSimpleDelay = new DelayFilter(delayLength, maxDelayLength, DelayFilter::LOAD_RANDOM);
  
  // Moving Average filter (acts as a low pass filter)
  m_pAvgFilter = new MovingAvgFilter(2);
  m_pAvgFilter->setGain(DECAY_FEEDBACK_GAIN);

  // Patch up the feedback loop
  m_pSimpleDelay->connectTo(m_pAvgFilter);
  m_pAvgFilter->connectTo(m_pSimpleDelay);

  m_numReleaseSamples = baseNumReleaseSamples*delayLength;
  m_SamplesSinceRelease = 0;

  // Store callback stuff
  m_releaseCallback = callback;
  m_callbackUserData = callbackUserData;

  //Initialize state
  m_state = STATE_ACTIVE;
}

PluckedString::~PluckedString(void)
{
  delete m_pSimpleDelay;
  delete m_pAvgFilter;
}

SAMPLE PluckedString::generateOutput(SAMPLE input)
{
  SAMPLE output;
  
  if (m_state == STATE_FINISHED) {
    printf("state finished\n");
    output = 0.0;
  }
  else {
    // Activate the audio chain
    m_pSimpleDelay->tick(0.0);
    output = m_pSimpleDelay->getOutput();
    m_pSimpleDelay->endOfTick();

    // Count samples since release event,
    if (m_state == STATE_RELEASED) {
      m_SamplesSinceRelease++;
    }

    // If enough samples have passed since the start of the
    // release phase, we call the callback to let the creator
    // of this object know that it is ready to be destroyed
    if ((m_SamplesSinceRelease > m_numReleaseSamples) && (m_state == STATE_RELEASED)) {
      m_state = STATE_FINISHED;

      if (m_releaseCallback != NULL) {
        m_releaseCallback(this, m_callbackUserData);
      }
    }
  }

    return output;
}

void PluckedString::release(void)
{
  m_state = STATE_RELEASED;
  m_pAvgFilter->setGain(RELEASE_FEEDBACK_GAIN);
}


int PluckedString::frequencyToDelayLength(MyFloat64 freq, int samplingRate)
{
  int roundedDelayLength; 
  MyFloat64 delayLength;
  // Calculate the needed delay for the Karplus-Strong algorithm, 
  // and create such a delay filter.
  delayLength = ((MyFloat64) samplingRate  - 0.5 * freq) / freq;
  
  // Use the nearest integer for the delay length
  roundedDelayLength = (int) ROUND__(delayLength);

  DEBUG_PRINTF("frequencyToDelayLength: freq= %f, delay = %d\n", freq, roundedDelayLength);
  
  return roundedDelayLength;
}

