#include <iostream>
#include <stdio.h>
#include "RtAudio.h"
#include "RtMidi.h"
#include <signal.h>
#include <math.h>
#include <assert.h>
#include  <stdlib.h>
#include <string.h>
#include <map>
#include <list>

// Project includes
#include "PolyCommon.h"
#include "DelayFilter.h"
#include "PassThroughElement.h"
#include "MovingAvgFilter.h"
#include "PluckedString.h"


/*********************************************************************************
                              Operating System Abstraction
**********************************************************************************/

// Operating system dependent thread functionality.
#if defined(__WINDOWS_DS__) || defined(__WINDOWS_ASIO__)
  #include <windows.h>
  #include <process.h>

  // Mutex related definition
  typedef CRITICAL_SECTION OsMutex;
  #define MUTEX_INITIALIZE(A) InitializeCriticalSection(A)
  #define MUTEX_DESTROY(A)    DeleteCriticalSection(A)
  #define MUTEX_LOCK(A)       EnterCriticalSection(A)
  #define MUTEX_UNLOCK(A)     LeaveCriticalSection(A)

  #define OS_SLEEP( X ) Sleep(X)
  #define RT_AUDI_API_TYPE     (RtAudio::WINDOWS_ASIO)

#elif defined(__LINUX_ALSA__) || defined(__UNIX_JACK__) || defined(__LINUX_OSS__) || defined(__MACOSX_CORE__)
  #include <unistd.h>  
  #include <pthread.h>
  
  // Mutex related definitions
  typedef pthread_mutex_t OsMutex;
  #define MUTEX_INITIALIZE(A) pthread_mutex_init(A, NULL)
  #define MUTEX_DESTROY(A)    pthread_mutex_destroy(A)
  #define MUTEX_LOCK(A)       pthread_mutex_lock(A)
  #define MUTEX_UNLOCK(A)     pthread_mutex_unlock(A)

  #define OS_SLEEP( X ) usleep(X)

#if defined (__LINUX_ALSA__)
  #define RT_AUDIO_API_TYPE     (RtAudio::LINUX_ALSA)
#elif defined (__UNIX_JACK__)
  #define RT_AUDIO_API_TYPE     (RtAudio::UNIX_JACK)
#else
  #error
#endif

#endif

using namespace std;

/*********************************************************************************
                                     Macros
**********************************************************************************/
// Audio device parameters
#define SAMPLE_RATE     (48000)
#define OUT_CHANNELS    (2)
#define IN_CHANNELS     (0)
#define BUFFER_FRAMES   (8)

// PI
#define MY_PI           (3.14159265358979)

// MIDI Definitions
#define MIDI_GET_MSG_TYPE(_statusByte)    ((_statusByte & 0xF0) >> 4)
#define MIDI_GET_CHANNEL(_statusByte)     (_statusByte & 0x0F)
#define MIDI_NOTE_OFF                     (8)
#define MIDI_NOTE_ON                      (9)
#define MIDI_NUM_NOTES                    (128)
#define MIDI_MIDDLE_A                     (69)

/*********************************************************************************
                                 Type Definitions
**********************************************************************************/
// Typedefs for iterators over an STL map and list
typedef map<unsigned int, PluckedString*>::iterator MapIterator;
typedef list<unsigned int>::iterator FinishedNotesIterator;

/*********************************************************************************
                                     Globals
**********************************************************************************/
// RtAudio - audio input and output
RtAudio * pAudio = NULL;

// RtMidiIn - MIDI input
RtMidiIn *pMidiIn = NULL;

// Mapping from MIDI note number to a pointer to the associated PluckedString
// object.
map<unsigned int, PluckedString*> noteMap;

// List of notes that have completed their release stage and need whose
// PluckedString objects need to be destroyed
list<unsigned int> finishedNotes;

// Mutex to protect our map (and its elements).
OsMutex mapMutex;

// Mapping of MIDI note number to frequency
MyFloat64 midiNoteToFreq[MIDI_NUM_NOTES];

/*********************************************************************************
                                  Forward Declarations
**********************************************************************************/
void stringDoneCallback( PluckedString* pString, unsigned int userData) ;

/*********************************************************************************
                                Function Implementations
**********************************************************************************/

////////////////////////////////////////////////////////////////
// Function Name: MidiInputCallback
// Description: 
// Parameters: 
// Returns:
////////////////////////////////////////////////////////////////
void MidiInputCallback( double timeStamp, std::vector<unsigned char> *message, void *userData)
{
  unsigned int msgSize = message->size();
  MyUint8 msgType;
  unsigned int noteNumber, noteVelocity;
  PluckedString* pPlucked = NULL, *pRemoveString = NULL;

  DEBUG_PRINTF("Midicallback: in\n");
  DEBUG_PRINTF("Got Midi message: \n");

  MIDI_DEBUG("---------------------------------------------\n");
  
  if((msgSize < 2) || (msgSize > 3)) {
    MIDI_DEBUG("Invalid Length. Ignoring MIDI message!\n");
  }
  else {
    msgType = MIDI_GET_MSG_TYPE(message->at(0));
    switch (msgType) {
    case MIDI_NOTE_ON:
    case MIDI_NOTE_OFF:
      if (msgSize != 3) {
        MIDI_DEBUG("Invalid Length. Ignoring MIDI message!\n");
        break;
      }
      
      noteNumber = (unsigned int) message->at(1);
      noteVelocity = (unsigned int) message->at(2);
      
      // Some devices send a note_on with velocity 0 to indicate a note_off.
      if ((noteVelocity == 0) && (msgType == MIDI_NOTE_ON)) {
      	msgType = MIDI_NOTE_OFF;
      }
      
      if (msgType == MIDI_NOTE_ON) {
        MIDI_DEBUG("Note ON: note number = %d, Note Velocity = %d\n",noteNumber, noteVelocity);
        /* Note On handling */

        // --- Start Critical Section ---
        MUTEX_LOCK(&mapMutex);
        DEBUG_PRINTF("Midi: LOCK\n");


        // Do we already have this note?
        if(noteMap.find(noteNumber) != noteMap.end()) {

          // If the note already exists, we delete the old note
          // and create a new one. This allows the player to hit notes 
          // rapidly, without having to wait for their release stage to
          // finish.

          // Look up the note
          MapIterator mapIterator = noteMap.find(noteNumber);
          pRemoveString  = mapIterator->second;

          // Destroy and remove String
          delete pRemoveString ;
          noteMap.erase(noteNumber);
        }

      	// Create new note
        pPlucked = new PluckedString(midiNoteToFreq[noteNumber], SAMPLE_RATE, stringDoneCallback, noteNumber);
        // Change the gain according to the velocity. 
        pPlucked->setGain((MyFloat64)noteVelocity / 127.0);
	
        // Insert to map
        noteMap.insert(pair<unsigned int,PluckedString*>(noteNumber,pPlucked));

      	DEBUG_PRINTF("Midi: UNLOCK\n");

        // --- End Critical Section ---
        MUTEX_UNLOCK(&mapMutex);       
      }
      else if (msgType == MIDI_NOTE_OFF) {

        MIDI_DEBUG("Note OFF: note number = %d, Note Velocity = %d\n",noteNumber, noteVelocity);

        /* Note off handling */

        // --- Start Critical Section ---
        MUTEX_LOCK(&mapMutex);
      	DEBUG_PRINTF("Midi: LOCK2\n");

        // Look up the note
        MapIterator mapIterator = noteMap.find(noteNumber);
        if (mapIterator != noteMap.end()) {
          // Note found
          pPlucked = mapIterator->second;

          // Notify note of release
          pPlucked->release();
        }        
	
        // --- End Critical Section ---
      	DEBUG_PRINTF("Midi: UNLOCK\n");
        MUTEX_UNLOCK(&mapMutex);

      }
      break;
    default:
      MIDI_DEBUG("MIDI message: 0x%x\n", msgType);
      MIDI_DEBUG("Unsupported message type. Ignoring MIDI message!\n");
      break;
    }
  }
  DEBUG_PRINTF("Midicallback: out\n");

}


////////////////////////////////////////////////////////////////
// Function Name: AudioCallback
// Description: 
// Parameters: 
// Returns:
////////////////////////////////////////////////////////////////
int AudioCallback( void *outputBuffer, void *inputBuffer,
                     unsigned int nFrames,
                     double streamTime,
                     RtAudioStreamStatus status,
                     void *userData )
{
  SAMPLE * pOutputSample = (SAMPLE *) outputBuffer;
  SAMPLE  output = 0.0;
  PluckedString* pCurrentString = NULL, *pRemoveString = NULL;

  DEBUG_PRINTF("AudioCallback: in\n");

  if (status == 0) {
    // Iterate over all samples
    for (unsigned int i=0; i<nFrames; i++) {
      output = 0.0;

      // --- Start Critical Section
      MUTEX_LOCK(&mapMutex);
      DEBUG_PRINTF("AudioCallback: LOCK\n");
      
      // Get output from active strings.
      for( MapIterator iter = noteMap.begin(); iter!=noteMap.end(); ++iter) {
        //cout << "Playing note #" << (*iter).first << endl;
        
        // Extract pointer to the string object
        pCurrentString = iter->second;

        // Standard sequence for interacting with an audio chain. In this case,
        // the audio chain is made of only one element - the string object.
        // See README for a description of the audio chain class hierarchy.
        pCurrentString->tick(0);
        output += pCurrentString->getOutput();
        pCurrentString->endOfTick();
      }

      // Delete finished Strings
      for( FinishedNotesIterator iter = finishedNotes.begin(); iter!= finishedNotes.end(); ++iter ) {
        unsigned int noteNumber = *iter;       
        // Look up the note
        MapIterator mapIterator = noteMap.find(noteNumber);

        // Make sure note is there
	assert(mapIterator != noteMap.end());
	pRemoveString  = mapIterator->second;
	  
	DEBUG_PRINTF("Deleting note %d\n", noteNumber);

	// Destroy and remove String
	int numErase = noteMap.erase(noteNumber);        

	delete pRemoveString ;	 
      }
      finishedNotes.clear();

      DEBUG_PRINTF("AudioCallback: UNLOCK\n");

      // --- End Critical Section
      MUTEX_UNLOCK(&mapMutex);


      for (int j = 0; j<OUT_CHANNELS; j++) {
        // write out output
        *pOutputSample = (output);

        // Advance output sample pointers
        pOutputSample++;
      }
    }
  }
  else {
    cerr << "Overflow/Underflow occured, status is " << status << endl;
  }

   DEBUG_PRINTF("AudioCallback: out\n");

  return 0;
}

////////////////////////////////////////////////////////////////
// Function Name: stringDoneCallback
// Description: A callback called by Plucked String objects once they
//              have determined that their post-release stage is finished,
//              and they are no longer generatign any useful audio.
// Parameters: 
// Returns:
////////////////////////////////////////////////////////////////
void stringDoneCallback( PluckedString* pString, unsigned int userData) 
{
    DEBUG_PRINTF("stringDoneCallback: in\n");
  // --- Start Critical Section ---
  //MUTEX_LOCK(&mapMutex); -- on unix this does not seem to acquire the lock even if the current thread already owns it
 
  // Add this note to the list of notes that need to removed from the map
  finishedNotes.push_back(userData);

  // --- End Critical Section ---
  //MUTEX_UNLOCK(&mapMutex);

  DEBUG_PRINTF("stringDoneCallback: out\n");
}

////////////////////////////////////////////////////////////////
// Function Name: cleanUp
// Description: Tries to close the audio stream and release the RtAudio object
//              
// Parameters: none
// Returns: void
////////////////////////////////////////////////////////////////
void cleanUp (void)
{
  DEBUG_PRINTF("cleanup: in\n");
  try {
    
    if (pAudio) {// cleanup RtAudio
      pAudio->closeStream();    
      delete pAudio;
      pAudio = NULL;
    }

    // cleanup RtMidi
    if (pMidiIn) {
      pMidiIn->cancelCallback();
      pMidiIn->closePort();
      delete pMidiIn;
      pMidiIn = NULL;
    }

    // Delete all strings
    for( MapIterator iter = noteMap.begin(); iter!=noteMap.end(); ++iter) {
      PluckedString* pCurrentString = iter->second;
      delete pCurrentString;
    }

  } catch (...) {
    cerr << "Exception Caught!" << endl;
    //err.printMessage();
  }
  DEBUG_PRINTF("stringDoneCallback: out\n");
}

////////////////////////////////////////////////////////////////
// Function Name: signalHandler
// Description: Cleans up and exits
// Parameters:
// Returns:
////////////////////////////////////////////////////////////////
void signalHandler(int sig)
{
  cleanUp();
  exit(1);
}


////////////////////////////////////////////////////////////////
// Function Name: calculateFrequencies
// Description: Build lookup table mapping midi notes to their frequency
// Parameters: 
// Returns:
////////////////////////////////////////////////////////////////
void calculateFrequencies(void)
{
  MyFloat64 halfToneDelta = pow(2.0,1.0/12.0);
  for(int i=0; i < MIDI_NUM_NOTES; i++) {
    midiNoteToFreq[i] = 440.0*pow(halfToneDelta, (MyFloat64)i - (MyFloat64)MIDI_MIDDLE_A);
    //cerr << "Note " << i << " mapped to frequency " << midiNoteToFreq[i] << endl;
  }
}

////////////////////////////////////////////////////////////////
// Function Name: main
// Description: Main thread for the program. Mostly responsible for kicking
//              RtAudio and RtMidi into action. After that spins in an infinite loop
//              while other threads do the work.
// Returns: 0 for success, 1 upon error.
////////////////////////////////////////////////////////////////
int main(int argc, char** argv) {  
  RtAudio::StreamParameters outputParameters, inputParameters;
  RtAudio::StreamParameters * pInParameters = NULL;
  unsigned int sampleRate = SAMPLE_RATE;
  unsigned int bufferFrames = BUFFER_FRAMES;
  int retVal = 0;
  
  // Register some signal handlers to allow clean exiting.
  // This was needed in Windows/Asio with a USB audio interface. When pressing Ctrl-C (or stoping the debugger)
  // the system would freeze (until you physically disconnect the USB cable...)
  // The solution was to implement some cleanup in these handlers.
  (void) signal(SIGINT, signalHandler);
  (void) signal(SIGABRT, signalHandler);
  (void) signal(SIGTERM, signalHandler);

  // Initialize the mutex responsible for protecting the map data structure
  MUTEX_INITIALIZE(&mapMutex);

  // Build midi-to-frequency table
  calculateFrequencies();

  // Greetings
  cerr << "Welcome! Please help yourself to as much polyphony as you can handle!" << endl;

  try { 
    pAudio = new RtAudio( RT_AUDIO_API_TYPE );
    pMidiIn = new RtMidiIn();

    // Initialize output parameters
#if defined(__WINDOWS_DS__) || defined(__WINDOWS_ASIO__)
    // hack for my Windows setup
    outputParameters.deviceId = 1;
#else // Linux or Mac
    outputParameters.deviceId = pAudio->getDefaultOutputDevice();
#endif
    outputParameters.firstChannel = 0;
    outputParameters.nChannels = OUT_CHANNELS;

    // Open the audio stream
    pAudio->openStream( &outputParameters,
      pInParameters,
      RTAUDIO_FLOAT64,
      sampleRate,
      &bufferFrames, 
      AudioCallback,
      NULL);

    // Open an input MIDI port
    if (pMidiIn->getPortCount() == 0) {
      cerr << "No Midi ports found. MIDI will not work. Continuing with audio only..." << endl;
    }
    else {
      pMidiIn->setCallback( &MidiInputCallback );
      // Open the MIDI port. The MIDI callback may be called immediately after this line.
      pMidiIn->openPort(1);
    }

    // Start the audio stream
    pAudio->startStream();  

    cout << endl << "Press 'q' and then Enter to quit" << endl;
    
    // Infinite loop to handle keyboard input. 
    // Note: The audio and MIDI are handled other threads.
    while( true ) {
      char inChar;
      cin >> inChar;
      if (inChar == 'q') {
        goto cleanup;
      }
      else {
        cerr << "Say what?!" << endl;
      }
      OS_SLEEP(1000);
    }
  } catch (RtError & err) {
    err.printMessage();
    retVal = 1;
    goto cleanup;
  }

cleanup:
  cleanUp();

  cerr << "Exiting...\n" << endl;  
  return 0;
}

